## This implementation is a rewrite of python codes on RFC8032 as ruby.
## original : https://tools.ietf.org/html/rfc8032

module Ed25519Keccak
  # base class of Ed25519
  class Ed25519Base
    def initialize( hash_function )
      @hash_function = hash_function
    end

    def sign(secret, message, form=:byte)
      arguments = change_arguments_format( [secret,message], form )
      return change_result_format( p_sign(*arguments), form)
    end

    def verify(public_key, message, signature, form=:byte)
      arguments = change_arguments_format( [public_key,message,signature], form )
      return p_verify(*arguments)
    end

    # calculate public key from secret
    def secret_to_public(secret, form=:byte)
      publickey = p_secret_to_public( change_argument_format(secret, form) )
      return change_result_format( publickey, form )
    end

    private
    ## region arguments/result util ##

    # Argument is always changed to a byte-string
    # if specified format is unknown, data is not changed.
    # format type is symbol
    def change_argument_format( argument , form )
      case form
      when :hex then
        return hexstr_to_bytes(argument)
      end
      # Other than those above : no change
      return argument
    end

    def change_arguments_format( arguments, form )
      result = []
      arguments.each{ |argument| result.push( change_argument_format( argument, form) ) }
      return result
    end

    # result_data is always byte-string
    def change_result_format( result_data, form )
      case form
      when :hex then
        return bytes_to_hexstr(result_data)
      end
      # Other than those above : no change
      return result_data
    end


    ## region common util ##

    def hexstr_to_bytes( hex_str )
      raise  ArgumentError , "hex-string length should be even" if hex_str.length.odd?
      return [hex_str].pack("H*")
    end

    def bytes_to_hexstr( binary )
      return binary.unpack("H*").first
    end

    # convert int to byte-string (littleEndian)
    # num    : unsigned number to convert to bytes
    # length : byte length
    def int_to_bytes( num , length )
      raise ArgumentError , "num :" + num.to_s if num < 0
      raise ArgumentError , "length :" + length.to_s if length < 0

      hex_str = num.to_s(16)
      hex_str = hex_str.rjust(length*2,"0")
      return hexstr_to_bytes(hex_str).reverse[0,length]
    end

    # convert byte-string to int (read bytes in little endian)
    def int_form_bytes(b)
      return bytes_to_hexstr(b.reverse).to_i(16)
    end


    ## region hash function ##

    def hash512(s)
      return @hash_function.call(s)
    end


    ## region calculate mod ##

    # calculate base**exponent % modulus
    def pow_mod(base, exponent, modulus)
      raise ArgumentError if exponent<0 || modulus<0
      # result of Nmod1 is always 0
      return 0 if modulus.equal? 1
      result = 1
      base = base % modulus
      while exponent > 0
        result = result*base%modulus if (exponent%2).equal? 1
        exponent = exponent >> 1
        base = base*base%modulus
      end
      return result
    end

    # calculate 1/x modp
    def modp_inv(x)
      return pow_mod(x , @@p-2 , @@p)
    end

    # calculate Hash(s) modp
    #
    def hash512_modq(s)
      return int_form_bytes(hash512(s)) % @@q
    end


    ## region calculate point of ed25519 curve ##

    # Points are represented as array [X, Y, Z, T] of extended
    # coordinates, with x = X/Z, y = Y/Z, x*y = T/Z
    #
    # [arguments]
    # pa : point A
    # pb : point B
    def point_add( pa , pb )
      _A = (pa[1]-pa[0]) * (pb[1]-pb[0]) % @@p
      _B = (pa[1]+pa[0]) * (pb[1]+pb[0]) % @@p
      _C = 2 * pa[3] * pb[3] * @@d % @@p
      _D = 2 * pa[2] * pb[2] % @@p
      _E = _B - _A
      _F = _D - _C
      _G = _D + _C
      _H = _B + _A
      return [ _E*_F , _G*_H , _F*_G , _E*_H ]
    end

    # Computes pointQ = s * pointA
    def point_mul(s, pa)
      pq = [0, 1, 1, 0]  # Neutral element
      while s > 0 do
        pq = point_add(pq, pa) unless (s & 1).equal? 0
        pa = point_add(pa, pa)
        s >>= 1
      end
      return pq
    end

    # return point A == point B
    def point_equal(pa, pb)
      # x1 / z1 == x2 / z2  <==>  x1 * z2 == x2 * z1
      return false if (pa[0] * pb[2] - pb[0] * pa[2]) % @@p != 0
      return false if (pa[1] * pb[2] - pb[1] * pa[2]) % @@p != 0
      return true
    end


    ## region point manipulation ##

    # Compute corresponding x-coordinate, with low bit corresponding to
    # sign, or return nil on failure
    def recover_x(y, sign)
      return nil if y >= @@p
      # x2 means x^2
      x2 = (y*y-1) * modp_inv(@@d*y*y+1)
      # when x2==0 and sign!=0, these combination of arguments is illegal
      if x2.equal? 0 then
        unless sign.equal? 0 then
          return nil
        else
          return 0
        end
      end
      # Compute square root of x2
      x = pow_mod(x2 , ((@@p+3) / 8) , @@p)
      x = x * @@modp_sqrt_m1 % @@p unless ((x*x - x2) % @@p).equal? 0
      return nil unless ((x*x - x2) % @@p).equal? 0
      x = @@p - x  unless (x & 1).equal? sign
      return x
    end

    # compress element[point] of ed25519 curve
    #
    # = Compressed format of Point(x,y) =
    # [ y0,y1...y253,y254 | x0 ] ArgumentError ,
    #
    # Change the most significant bit of "y" to the least significant bit of "x"
    # The most significant bit of "y" is always zero because "0 <= y < 2^255-19"
    def point_compress(p)
      zinv = modp_inv(p[2])
      x = p[0] * zinv % @@p
      y = p[1] * zinv % @@p
      # OR can be used to set the least significant bit of x to the most significant bit of y
      # because the most significant bit of "y" is always zero
      c = y | ((x & 1) << 255)
      return int_to_bytes(y | ((x & 1) << 255) , 32)
    end

    # decompress point that is compressed into 32bytes
    def point_decompress(s)
      # check argument
      raise ArgumentError , "Invalid input length for decompression" unless s.length.equal? 32

      y = int_form_bytes(s)
      sign = y >> 255
      y &= (1 << 255) - 1
      x = recover_x(y, sign)
      if x.nil? then
        return nil
      else
        return [x, y, 1, x*y % @@p]
      end
    end


    ## region key manipulation ##

    # hash512(secret)
    # => HASH(512bit)
    # => [LH(256bit)] / [RH(256bit)]
    # => LH -> (set some bits) -> a
    # return ( a , RH )
    def secret_expand(secret)
      raise "Bad size of private key" unless secret.length.equal? 32

      h = hash512(secret)
      a = int_form_bytes(h[0,32])
      a &= (1 << 254) - 8
      a |= (1 << 254)
      return [a, h[32,32]]
    end


    ## region sign and verify ##

    # [ signature format ]
    # | compressed data of _R | s | <- concatnate
    def p_sign(secret, message)
      a, prefix = secret_expand(secret)
      _A = point_compress(point_mul(a, @@G))
      r = hash512_modq(prefix + message)
      _R = point_mul(r, @@G)
      _Rs = point_compress(_R)
      h = hash512_modq(_Rs + _A + message)
      s = (r + h * a) % @@q
      return _Rs + int_to_bytes(s,32)
    end

    def p_verify(public_key, message, signature)
      # check arguments
      raise  ArgumentError , "Bad public_key key length" unless public_key.length.equal? 32
      raise  ArgumentError , "Bad signature length" unless signature.length.equal? 64

      _A = point_decompress(public_key)
      return false if _A.nil? || _A.equal?(0)

      _Rs = signature[0,32]
      _R = point_decompress(_Rs)
      return false if _R.nil? || _R.equal?(0)

      s = int_form_bytes(signature[32,32])
      return false if s >= @@q

      h = hash512_modq(_Rs + public_key + message)
      sB = point_mul(s, @@G)
      hA = point_mul(h, _A)
      return point_equal(sB, point_add(_R, hA))
    end

    # public_keyKey = aG
    # "a" is generated form a secret
    def p_secret_to_public(secret)
      expanded = secret_expand(secret)
      a = expanded.first
      return point_compress(point_mul(a, @@G))
    end


    ## region variable ##

    # Base field Z_p
    @@p = 2**255 - 19
    # Curve constant
    # d = -121665 * modp_inv(121666) % p
    @@d = 37095705934669439343138083508754565189542113879843219016388785533085940283555
    # Group order
    @@q = 2**252 + 27742317777372353535851937790883648493
    # Square root of -1
    # modp_sqrt_m1 = pow_mod(2 , ((p-1) / 4) , p)
    @@modp_sqrt_m1 = 19681161376707505956807079304988542015446066515923890162744021073123829784752
    # Base point
    # g_y = 4 * modp_inv(5) % p
    # g_x = recover_x(g_y, 0)
    @@g_y = 46316835694926478169428394003475163141307993866256225615783033603165251855960
    @@g_x = 15112221349535400772501151409588531511454012693041857206046113283949847762202
    @@G   = [ @@g_x, @@g_y, 1, @@g_x * @@g_y % @@p ]
  end
end
